package nstaclient

import (
	"context"
	"fmt"
	"log"
	"os"
	"os/signal"
	"path/filepath"
	"syscall"
	"time"

	"github.com/it-novum/nsta/checkmk"
	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpproto"
	"github.com/it-novum/nsta/httpserver"
	"github.com/it-novum/nsta/nstautil"
	"github.com/it-novum/nsta/satstatus"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

type ClientPull struct {
	ClientSettings

	gearman          *gearman.Client
	webserver        *httpserver.NSTAHTTPServer
	nagiosConfigPath string
	toLocalSat       gearman.SyncMessageFIFO
	toMaster         gearman.SyncMessageFIFO
	cmk              checkmk.Checkmk
	cmdBackend       *checkmk.LocalCommandBackend

	lastExport time.Time
}

func (c *ClientPull) SatelliteID() int {
	return c.ClientSettings.SatelliteID
}

func (c *ClientPull) SatelliteName() string {
	return "Local"
}

func (c *ClientPull) APIKey() string {
	return c.ClientSettings.APIKey
}

// Get called by the webserver - stores messages recived from the master server
func (c *ClientPull) Incoming(msg *gearman.Message) {
	c.toLocalSat.Push(msg)
}

// Get called by the webserver - returns all messages the satellite wants to send to the master server
func (c *ClientPull) Outgoing() []*gearman.Message {
	return c.toMaster.GetAll(8192)
}

func (c *ClientPull) LastExport() time.Time {
	return c.lastExport
}

func (c *ClientPull) Export(msg *httpproto.ExportData) error {
	err := nstautil.InstallExport(c.nagiosConfigPath, msg.TARBase64)
	if err == nil {
		c.lastExport = msg.LastExport
	}
	return err
}

// WebStatus is only required on the master, but implements SatelliteIO
func (c *ClientPull) WebStatus(_ *satstatus.Status) {
}

// In Pull mode, the openITCOCKPIT Main Server pulls the check result from the Satellite System
func (c *ClientPull) StartForeground() error {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	c.lastExport = time.Time{}
	c.nagiosConfigPath = filepath.Join(c.NagiosPath, "etc")
	c.cmdBackend = &checkmk.LocalCommandBackend{
		Timeout:       time.Second * 120,
		SatelliteName: "Local",
	}

	c.webserver = httpserver.New(c.AccessLogWriter, c.NagiosPath, c.WebserverAddressHTTP, c.WebserverAddressHTTPS, c.CertFile, c.KeyFile, httpserver.WebserverTypeSatellite)

	c.cmk.SatelliteName = "Local"
	c.toLocalSat.Start()
	c.toMaster.Start()
	c.gearman = &gearman.Client{
		ClientOptions: gearman.ClientOptions{
			ClientType:     gearman.Satellite,
			ClientName:     "Local",
			Address:        c.GearmanAddress,
			MessageReceive: c.toMaster.Direct(),
		},
	}
	c.webserver.RegisterSatelliteIO(c)

	c.gearman.Start(ctx)
	c.webserver.Run()

	c.run(ctx)

	return nil
}

func (c *ClientPull) run(ctx context.Context) {
	defer func() {
		c.webserver.Stop()
		c.webserver.UnregisterSatelliteIO(c.SatelliteID())
		c.toLocalSat.Close()
		c.toMaster.Close()
	}()

	// Bind linux kernel signal handler
	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM)

	for {
		select {
		case <-ctx.Done():
			return
		case <-sigs:
			// Exit on SIGINT / SIGTERM
			return
		case <-c.toLocalSat.Notify():
			// Got notify signal from FIFO (message from the main system)
			msg := c.toLocalSat.Get()
			if msg != nil {
				switch msg.Type {
				case gearman.MessageCheckmk:
					// READ!! https://stackoverflow.com/a/30183893
					go func(gearmanMessageCopy *gearman.Message) {
						response, err := c.cmk.ProcessMessage(c.cmdBackend, msg).Serialize()
						if err != nil {
							log.Println("could no serialize checmk answer:", err)
						}

						//Send Checkmk result back to main system
						c.toMaster.Push(&gearman.Message{
							Type:   gearman.MessageCheckmk,
							String: response,
						})
					}(msg)
				case gearman.MessageTimesync:
					go func() {
						ts := gearman.TimesyncMessage{}
						if err := json.Unmarshal([]byte(msg.String), &ts); err != nil {
							log.Println(&Error{Name: "Local", Message: "received invalid timesync message", Err: err})
						} else {
							if _, err := c.cmdBackend.RunLocalCommand(ctx, "", []string{"sudo", "date", "-s", fmt.Sprint("@", ts.SetTime.Unix())}, map[string]string{}); err != nil {
								log.Println(err)
							} else {
								log.Println(&Error{Name: "Local", Message: "successful updated time"})
							}
						}
					}()
				default:
					c.gearman.SendMessage(msg)
				}
			}
		}
	}
}
